#if 0
#include <opencv2/opencv.hpp>
#include <iostream>
#include "common.h"
#include <omp.h>
#include "tdrocr.h"
#include "utils.h"

/**
 * @brief 判断第二行mrz是否越出图像下边
 *        note：函数误判的可能性较低，只有当字符离边界很近时，并且背景底纹非常明显时才会误判
 *        代码逻辑:  step① 首先判断旋转矩形框是否越过边界超过3个像素，超过则return -1
 *                  step② 针对二代设备有立柱的情况，判断mrz下方是否有一条黑色的直线，有的话且大于阈值返回-1
 *                  step③ 求出在旋转矩形内部的图像最底行的一行像素，如果字符越出边界，则这一行图像二值化后黑色像素较多，黑色像素个数超过阈值th则认为越出边界
 * @param src       roi原图
 * @param vertices  外接旋转矩形的四个顶点坐标  
 * @param angle     mrz第二行的外接旋转矩形的angle
 * @param th        精度阈值1：越小越容易判断越出边界
 * @param offset_th 精度阈值2：越大越容易判断越出边界
 * @param plus      默认false，是否开启增强判断，判断立柱情况，有可能误判
 * @return int      没越出边界返回0，越出边界返回负数
 */
int is_outBottom(Mat& src,  Size rect_size, Point2f center, float angle, int th=9, int offset_th=3, bool plus=false){

    int boundary;
    RotatedRect rotatedRect(center, rect_size, angle);
    Point2f vertices[4];
    rotatedRect.points(vertices);

    // 求出最低点 和 最低点相邻的两个顶点
    Point2f Lowest_point;
    Point2f Adj_p1, Adj_p2;
    
    if (angle < 0)
    {
        Lowest_point = vertices[0];
        Adj_p1 = vertices[1];
        Adj_p2 = vertices[3];
    }
    else
    {
        Lowest_point = vertices[3];
        Adj_p1 = vertices[2];
        Adj_p2 = vertices[0];
    }
    
    int img_h = src.rows;
    float dis = Lowest_point.y - img_h;
    if(dis > 3)
        return -1;

    if(plus){
        if(dis < 5) // 二代设备有立柱的情况，粗略筛选一下
        {
            if (abs(angle) < 1)
            {
                int last_line = Lowest_point.y;
                if(last_line+6>=img_h)
                    last_line = img_h - 7;
                for (int i = -1; i < 2; i++)
                {
                    Mat src_lastRow = src.row(last_line + i*3);
                    float x1, x2;
                    if (Adj_p1.x < Adj_p2.x)
                    {
                        x1 = Adj_p1.x;
                        x2 = Adj_p2.x;
                    }else{
                        x2 = Adj_p1.x;
                        x1 = Adj_p2.x;
                    }
                    float t = (x2 - x1)*0.6;
                    Mat lastRow = src_lastRow.colRange(cv::Range(x1, x2 + 1)).clone();
                    // 二值化，边框是黑色的
                    if(lastRow.empty()){
                        return 0;   // 左右越界会出现这种情况，所以此处返回0
                    }
                    cv::cvtColor(lastRow, lastRow, CV_BGR2GRAY);
                    cv::threshold(lastRow, lastRow, 50, 255, THRESH_BINARY_INV);
                    // 统计黑色像素的个数
                    int zeroCount = countNonZero(lastRow);
                    if (zeroCount > t)
                    {
                        return -1;
                    }
                }
            }
        }

        if(dis < 0)
            return 0;
        // 求出旋转矩形两条边与src交界处的x坐标
        float x1, x2;
        if(angle==0){
            x1 = Adj_p1.x;
            x2 = Adj_p2.x;
        }
        else{
            if (Adj_p1.x - Lowest_point.x != 0) { // 防止除以零
                x1 = ((img_h-1 - Lowest_point.y) * (Adj_p1.x - Lowest_point.x)) / (Adj_p1.y - Lowest_point.y) + Lowest_point.x;
            }
            if (Adj_p2.x - Lowest_point.x != 0) { // 防止除以零
                x2 = ((img_h-1 - Lowest_point.y) * (Adj_p2.x - Lowest_point.x)) / (Adj_p2.y - Lowest_point.y) + Lowest_point.x;
            }   
        }

        if(x1>x2){
            float tmp = x1;
            x1 = x2;
            x2 = tmp;
        }
        // 确保x1和x2在图像宽度范围内
        if(x1 < 0)
            x1 = 0;
        if(x2 >= src.cols)
            x2 = src.cols - 1;

        // 从lastRow中提取x1到x2的列
        Mat src_lastRow = src.row(src.rows - 1 - offset_th);  // offset_th: 一般字符周围颜色不准确，往上偏移几行的结果更准确;
        Mat lastRow = src_lastRow.colRange(cv::Range(x1, x2 + 1)).clone();
        // line(src, Point2f(x1, src.rows - 1 - offset_th), Point2f(x2 + 1, src.rows - 1 - offset_th), Scalar(0, 255, 2));
        // imshow("SSSSS", src);
        // waitKey(0);
        // imwrite("./bIiimg.png", lastRow);

        // 二值化，字符是黑色的
        if(lastRow.empty()){
            return 0;   // 左右越界会出现这种情况，所以此处返回0
        }
        cvtColor(lastRow, lastRow, CV_BGR2GRAY);
        threshold(lastRow, lastRow, 37, 255, THRESH_BINARY_INV);
        // 统计黑色像素的个数
        int zeroCount = countNonZero(lastRow);
        // cout << "zeroCount: " << zeroCount << endl;
        // cout << "th: " << th << endl;

        if (zeroCount > th)
        {
            return -1;
        }
    }
    return 0;

}

/**
 * @brief 判断第1行mrz是否越出图像上边界
 *        note：函数误判的可能性较低，只有当字符离边界很近时，并且背景底纹非常明显时才会误判
 *        代码逻辑:  step① 首先判断旋转矩形框是否越过边界超过3个像素，不超过则return 0，超过则step2
 *                  step② 求出在旋转矩形内部的图像最顶行的一行像素，如果字符越出边界，则这一行图像二值化后黑色像素较多，黑色像素个数超过阈值th则认为越出边界
 * @param src       roi原图
 * @param vertices  外接旋转矩形的四个顶点坐标  
 * @param angle     mrz第1行的外接旋转矩形的angle
 * @param th        精度阈值1：越小越容易判断越出边界
 * @param offset_th 精度阈值2：越大越容易判断越出边界
 * @return int      没越出边界返回0，越出边界返回负数
 */
int is_outTop(Mat& src,  Size rect_size, Point2f center, float angle){
     // 求出最高点 和 最高点相邻的两个顶点
    Point2f highest_point;
    Point2f Adj_p1, Adj_p2;
    RotatedRect rotatedRect(center, rect_size, angle);
    Point2f vertices[4];
    rotatedRect.points(vertices);
    if (angle < 0)
    {
        highest_point = vertices[2];
        Adj_p1 = vertices[1];
        Adj_p2 = vertices[3];
    }
    else
    {
        highest_point = vertices[1];
        Adj_p1 = vertices[2];
        Adj_p2 = vertices[0];
    }

    // 外接矩形框没出界或越界很小，此时不会影响字符识别
    int img_h = src.rows;
    if (highest_point.y < 1)
        return -1;

    return 0;
}


/**
 * @brief 判断第mrz是否越出图像左右边界
 * 
 * @param src       roi原图
 * @param rect_size 外接旋转矩形的大小Size
 * @param center    外接旋转矩形位于ROI图上的中心点
 * @param angle     mrz第1行的外接旋转矩形的angle
 * @param plus      默认false, 是否增强判断，判断立柱情况，有可能误判
 * @return int 没越出边界返回0，越出左边界返回-1，越出右边界返回-2
 */
int is_outLeftRight(Mat& src, Size rect_size, Point2f center, float angle, bool plus=false){

    RotatedRect rotatedRect(center, rect_size, angle);
    Point2f vertices[4];
    rotatedRect.points(vertices);

    Point2f left_point, right_point;
    // Mat aa = drawRotatedRect(src, rect_size, center, angle);
    // imshow("aa", aa);
    // waitKey(0);
    // for(int i=0; i<4; i++){
    //     cout << vertices[i] << endl;
    // }
    // cout << "src.cols - 1: "<< src.cols - 1 << endl;
    if (angle < 0)
    {
        left_point  = vertices[1];
        right_point = vertices[3];
    }
    else
    {
        left_point  = vertices[0];
        right_point = vertices[2];
    }

    // float a_char_w = (rect_size.width / 45)/2 ;  // 算上字段前后两个空白


    // if (left_point.x < a_char_w)
    if (left_point.x < 0)
        return -1;
    if(right_point.x > src.cols-1)
        return -2;

    if(plus){
        Mat edge, biedge;
        float th = rect_size.height / 3;
        float src_center = src.cols / 2;

        // 判断左右立柱的情况
        for (int i = 1; i < 4; i++)
        {
        int zeroCount = 0;
        if (center.x > src_center)
        {
            Point2f v = right_point;
            v.x += i*2;
            Point2f p1;
            if (angle <= 0)
                p1 = Point2f(v.x + 1, v.y - rect_size.height - 1);
            else
                p1 = Point2f(v.x + 1, v.y + rect_size.height + 1);
                        
            if (v.x < 0)
                v.x = 0;
            else if (v.x > src.cols - 1)
                v.x = src.cols - 1;
            if (v.y < 0)
                v.y = 0;
            else if (v.y > src.rows - 1)
            {
                v.y = src.rows - 1;
            }
            if (p1.x < 0)
                p1.x = 0;
            else if (p1.x > src.cols - 1)
                p1.x = src.cols - 1;
            if (p1.y < 0)
                p1.y = 0;
            else if (p1.y > src.rows - 1)
            {
                p1.y = src.rows - 1;
            }
            edge = src(Rect(v, p1));
            // line(aa, v, p1, Scalar(0, 0, 255)); 
        }
        else
        {
            Point2f v = left_point;
            v.x -= 2*i;
            Point2f p1;
            if (angle >= 0)
                p1 = Point2f(v.x + 1, v.y - rect_size.height - 1);
            else
                p1 = Point2f(v.x + 1, v.y + rect_size.height + 1);
            
            if (v.x < 0)
                v.x = 0;
            else if (v.x > src.cols - 1)
                v.x = src.cols - 1;
            if (v.y < 0)
                v.y = 0;
            else if (v.y > src.rows - 1)
            {
                v.y = src.rows - 1;
            }
            if (p1.x < 0)
                p1.x = 0;
            else if (p1.x > src.cols - 1)
                p1.x = src.cols - 1;
            if (p1.y < 0)
                p1.y = 0;
            else if (p1.y > src.rows - 1)
            {
                p1.y = src.rows - 1;
            }
            edge = src(Rect(v, p1));
            // line(aa, v, p1, Scalar(0, 0, 255));
        }
        // imshow("aaa", aa);
        // waitKey(0);
        if(edge.empty()){
            if(center.x > src_center){
                return -2;
            }else{
                return -1;
            }
        }
        cvtColor(edge, biedge, CV_BGR2GRAY);
        threshold(biedge, biedge, 35, 255, THRESH_BINARY_INV);
        
        // adaptiveThreshold(edge, biedge, 255, ADAPTIVE_THRESH_MEAN_C, THRESH_BINARY_INV, 5, 0);
        // waitKey(0);
        zeroCount += countNonZero(biedge);
        if (zeroCount > th)
        {
            if(center.x > src_center){
                return -2;
            }else{
                return -1;
            }
        }
    }
    }
    return 0;
}


/**
 * @brief           判断mrz是否越出图像
 *                  note：函数误判的可能性较低，只有当字符离边界很近时，并且背景底纹非常明显时才会误判
 *                        无法处理二代设备，或图像四周有立柱的情况
 * @param src        一行mrz的外接矩形图
 * @param roi        roi原图
 * @param rect_size  mrz第二行的外接旋转矩形的Size
 * @param center     mrz第二行的外接旋转矩形的center
 * @param angle      mrz第二行的外接旋转矩形的angle
 * @param ori_center rect在原图(roi图)上的center
 * @param is_line1   用于区分mrz第一行还是第二行的标志
 * @param th         精度阈值1：越小越容易判断越出边界
 * @param offset_th  精度阈值2：越大越容易判断越出边界
 * @return int       没越出边界返回0，越出边界返回负数
 */
int is_outBound(Mat& src, Mat& roi, Size rect_size, Point2f center, float angle, Point2f ori_center, bool is_line1, int th=9, int offset_th=3){

    RotatedRect rotatedRect(center, rect_size, angle);
    Point2f vertices[4];
    rotatedRect.points(vertices);
    // Mat aa = drawRotatedRect(src, rect_size, center, angle);
    // imshow("rect_size", aa);
    if (is_line1)
    {
        int res_t = is_outTop(roi,rect_size, ori_center, angle);  // 上
        if(res_t<0){
            LOGD("Attention: Characters close to the upper boundary of the image");
            return ERROR_CHAR_TOP_IMG;
        }
    }
    else{
        int res_b = is_outBottom(roi, rect_size, ori_center, angle, th, offset_th);  // 下
        if(res_b<0){
            LOGD("Attention: Characters close to the lower boundary of the image");
            return ERROR_CHAR_BOTTOM_IMG;
        }
    }
    int res_lr = is_outLeftRight(roi, rect_size, ori_center, angle);
    if(res_lr==-1){
        LOGD("Attention: Characters close to the left boundary of the image");
        return ERROR_CHAR_LEFT_IMG;                   // 左
    }                               
    if(res_lr==-2){        
        LOGD("Attention: Characters close to the right boundary of the image");             
        return ERROR_CHAR_RIGHT_IMG;                   // 右
    }   
    return 0;
}

#endif
