#include <omp.h>
#include <opencv2/core/core.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <vector>
#include "card_type.h"
#include "common.h"
#include "passport_visa.h"
#include "result.h"
#include "tdr_face_detect.h"
#include "tdrocr.h"

const char* g_version_information = "tdrocr v3.7.0c-20240204";

void setTotalCount(int width, int height, int oldcamera);
void unload_model_centerface();
int getPassportFaceImage(cv::Mat& src, cv::Mat& faceimg);
int get_visa_issue_info(cv::Mat& src, OCR_RESULT& ocr_result, bool is_baseRect = false, int num_chars = 44);
int tdrocr_scan_passport(cv::Mat& src, OCR_RESULT& ocr_result);
void clearVisaIssueInfo();
int initVisaIssueInfo(char* path);
void uninit_crnn();
int isNewCamera();
int g_init_ok = 0;
extern int g_scan_none;
int init_card_ai(char * dir);
void uninit_card_ai();
extern cv::Mat g_faceimg, g_cropimg;
void setVisaCategory(OCR_RESULT & ocr_result);
int passportMatchOcr3(cv::Mat & line1_rgb, cv::Mat & line2_rgb, OCR_RESULT &ocr_result, int & sz1, int &sz2);
extern int g_scan_type;
int tdr_passport_scan(Mat& src, OCR_RESULT& ocr_result);
int tdr_mixed_scan(Mat& src, OCR_RESULT& ocr_result);
int tdr_mixed_scan_refine(Mat& src, OCR_RESULT& ocr_result);
void clearDocType();    
void tdrocr_reset();
int init_ncode_map();

vector<RotatedRect>  mrz_rotatedRect;  // 存放两行mrz的旋转外接矩形，用于最终裁切证件图像时定位用

void tdrocr_unInit() {
    match_unInit();
    unload_model_centerface();
    uninit_crnn();
    g_init_ok = 0;
    clearVisaIssueInfo();
    uninit_card_ai();
}

int tdrocr_init(const char* dir, const char* opt, int isNewCamera) {
    int ret;
    std::string model_dir = dir;
    LOGI("version: %s", g_version_information);

    if (isNewCamera == 0) {
        LOGI(">>>> isOldCamera <<<<\n");
        setTotalCount(24, 34, 0);
    } else if (isNewCamera == 1) {
        LOGI(">>>> isNewCamera <<<<\n");
        setTotalCount(19, 28, 1);
    } else {
        LOGI("tdrocr_init: isNewCamera=%d is not support", isNewCamera);
        return -1;
    }
    init_ncode_map();
    int model_opt=0;
    if(strcasecmp(opt, "opencl")==0){
         model_opt=1;
    }
    ret = init_crnn((char *)dir, model_opt);
    if (ret != 0) {
        LOGI("===> init1 fail <===  %d", ret);
        return ret;
    }

    LOGI("init 1 pass.");

    ret = match_init(dir);

    if (ret != 0) {
        LOGI("===> init2 fail <===  %d", ret);
        return ret;
    }

    LOGI("init 2 pass.");

    ret = tdr_face_init2(model_dir.c_str());
    if (ret != 0) {
        LOGI("===> init3 fail <===  %d", ret);
        return ret;
    }

    LOGI("init 3 pass.");
#ifdef USE_ISSUER_ADDR
    ret = initVisaIssueInfo((char*)dir);
    if (ret != 0) {
        LOGI("===> init4 fail <===  %d", ret);
        return ret;
    }
    LOGI("init 4 pass.");
#endif

    ret=init_card_ai((char *)dir);
    if (ret != 0) {
        LOGI("===> init5 fail <===  %d", ret);
        return ret;
    }
    LOGI("init 5 pass.");


    LOGI(">>>> init success <<<<");

    memset((char *)&g_next_frame, 0x00, sizeof(g_next_frame));

    g_init_ok = 1;
    return 0;
}

int tdrocr_scan_yuv(unsigned char* src_data, int width, int height, int frm_wd,
                    int frm_hg, int mode, OCR_RESULT& ocr_result) {
    clearDocType();
    if (g_init_ok != 1) {
        clearOcrResult(ocr_result);
        LOGI("tdrocr_init first!");
        return -1;
    }
    int ret = 0;
    cv::Mat mBgr;
    cv::Mat src(height + height / 2, width, CV_8UC1, (unsigned char*)src_data);
    cv::cvtColor(src, mBgr, CV_YUV2BGR_NV21);
    ret = tdrocr_scan_mat(mBgr, mode, ocr_result);

    return ret;
}

int tdrocr_scan_mat(cv::Mat& src, int mode, OCR_RESULT& ocr_result) {
    clearDocType();
    clearOcrResult(ocr_result);
    if (g_init_ok != 1) {
        LOGI("tdrocr_init first!");
        LOGI("ocr return = %d", -1);
        return -1;
    }
    cv::Mat dst;
    if (src.cols != 1600) {
        cv::Mat img_resize;
        double scale = src.cols / 1600.0;
        int width = round(src.cols / scale);
        int height = round(src.rows / scale);
        cv::resize(src, img_resize, Size(width, height), cv::INTER_NEAREST);
        dst = img_resize;
    } else {
        dst = src;
    }

    if (dst.rows < 800) {
        LOGI("image size invalid.");
        LOGI("ocr return = %d", ERROR_IMAGE_SIZE);
        return ERROR_IMAGE_SIZE;
    }

    g_cropimg.release();
    g_faceimg.release();
    int ret = tdrocr_scan_passport(dst, ocr_result);
    if(ret==0){
         g_next_frame.prev=0;
         g_detect_mode=0;
    }
    LOGI("ocr return = %d", ret);
    return ret;
}

int tdrocr_scan_passport(cv::Mat& src, OCR_RESULT& ocr_result) {
    if (g_init_ok != 1) {
        LOGI("tdrocr_init first!");
        return -1;
    }
#ifndef TDR_TRACE
    if (src.cols != 1600) {
        LOGI("image resolution %d X %d is not support!", src.cols, src.rows);
        return ERROR_IMAGE_SIZE;
    }
#endif
    tdrocr_reset();
    int ret;
    switch (g_scan_type) {
        case CERTIFICATE_TYPE_PASSPORT:  // 1 护照
        case CERTIFICATE_TYPE_VISA:      // 7 签证
        case CERTIFICATE_TYPE_RESIDENT:  // 8 外国人居留许可
        case CERTIFICATE_TYPE_ENTRY_EXIT_PERMIT:
        case CERTIFICATE_TYPE_SEAFARER:
        case CERTIFICATE_TYPE_FROM_TW_ONEOFF:  // 4 台湾来内地
            ret = tdr_passport_scan(src, ocr_result);
            break;
        case CERTIFICATE_TYPE_FROM_HM:  // 3 香港来内地
        case CERTIFICATE_TYPE_FROM_TW:  // 4 台湾来内地
        case CERTIFICATE_TYPE_APEC:
            ret = tdr_card_scan(src, ocr_result);
            break;
        case 99:
            ret = tdr_mixed_scan_refine(src, ocr_result);
            break;
        default:
            ret = tdr_mixed_scan(src, ocr_result);
            break;
      }

      return ret;
}

void tdrocr_set_scan_type(int scan_type) {
    switch (scan_type) {
        case CERTIFICATE_TYPE_DEFAULT:   // 0
        case CERTIFICATE_TYPE_PASSPORT:  // 1 护照
        // case CERTIFICATE_TYPE_ID:        // 2
        case CERTIFICATE_TYPE_FROM_HM:  // 3 香港来内地
        case CERTIFICATE_TYPE_FROM_TW:  // 4 台湾来内地
        case CERTIFICATE_TYPE_FROM_TW_ONEOFF:  // 4 台湾来内地

        // case CERTIFICATE_TYPE_TO_HM:     // 5
        // case CERTIFICATE_TYPE_TO_TW:     // 6
        case CERTIFICATE_TYPE_VISA:      // 7 签证
        case CERTIFICATE_TYPE_RESIDENT:  // 8 外国人居留许可
        case CERTIFICATE_TYPE_ENTRY_EXIT_PERMIT:
        case CERTIFICATE_TYPE_APEC:
        case CERTIFICATE_TYPE_SEAFARER:
            g_scan_type = scan_type;
            break;
        case 99:
            g_scan_type = 99;
            break;
        default:
            g_scan_type = CERTIFICATE_TYPE_DEFAULT;
            break;
      }

      LOGD("set_scan_type = %d", scan_type);
      g_detect_mode=0;
}

/**
 * @brief                   裁切护照、签证区域的图像。
 *                          根据mrz的位置与长度计算输出图像的rect。
 *  NOTE:                   字符个数取44时，可通用44个字符与36个字符的证件，只是在36字符证件上不够精确。
 * @param src               原图
 * @param dst               输出裁切后的签证护照图像
 * @param mrz_rotatedRect   存放两个mrz的数组
 * @param num_chars         字符个数，默认44个
 */
void cropImage(Mat &src, Mat&dst, vector<RotatedRect>& mrz_rotatedRect, int num_chars=44){
    Mat cropimg;
    if (mrz_rotatedRect.size() < 2) // 保护措施，注释掉该if后测试结果也没问题
    {   
        if (isNewCamera() == 1) {
            cropimg = src(Rect(120, 50, src.cols - 120 * 2, src.rows - 50 * 2));
        } else {
            cropimg = src(Rect(40, 40, src.cols - 40 * 2, src.rows - 40 * 2));
        }
    }
    else
    {
        float min_x, min_y, max_x, max_y;
        for (int i = 0; i < 2; i++)
        {
            cv::Point2f vertices[4];
            mrz_rotatedRect[i].center.y += (src.rows - CUT_SIZE - 1);
            mrz_rotatedRect[i].points(vertices);
            min_x = vertices[0].x;
            max_x = vertices[0].x;
            min_y = vertices[0].y;
            max_y = vertices[0].y;
            for (int j = 1; j < 4; j++)
            {
                if(vertices[j].x < min_x) min_x = vertices[j].x;
                if(vertices[j].x > max_x) max_x = vertices[j].x;
                if(vertices[j].y < min_y) min_y = vertices[j].y;
                if(vertices[j].y > max_y) max_y = vertices[j].y;
            }
        }

        // 估计一个字符的宽度，作为基准长度单位
        float r_len1 = max(mrz_rotatedRect[0].size.width, mrz_rotatedRect[0].size.height);
        float r_len2 = max(mrz_rotatedRect[1].size.width, mrz_rotatedRect[1].size.height);
        float a_char_w = (r_len1 / num_chars + r_len2 / num_chars) / 2;

        Point2f crop_p1, crop_p2;
        if (num_chars != 36)        // 非36个字符的护照均按44个算
        {                                                                                
            crop_p1 = Point2f(max(0.0f, min_x-a_char_w*3), max(0.0f, min_y-a_char_w*30));                               // 左上角的点
            crop_p2 = Point2f(min(src.cols - 1.0f, max_x + a_char_w * 3), min(src.rows - 1.0f, max_y + a_char_w * 3));  // 右下角
        }
        else                         // 36个字符，此处只是更精确，44个字符的分支可以适配36字符
        {                                                                                
            crop_p1 = Point2f(max(0.0f, min_x-a_char_w*2), max(0.0f, min_y-a_char_w*27));                               // 左上角的点
            crop_p2 = Point2f(min(src.cols - 1.0f, max_x + a_char_w * 4), min(src.rows - 1.0f, max_y + a_char_w * 2)); // 右下角
        }

        // 大角度时需要偏移一些像素
        float angle = mrz_rotatedRect[0].angle;
        if (angle < -45)
            angle += 90;
        float ar = sin(angle * CV_PI / 180.0);
        float rotated_offset = ar*a_char_w*35;
        if (angle > 2)
        {
            crop_p2.x = min(crop_p2.x + rotated_offset, src.cols- 1.0f);
        }
        else if(angle<-2){
            crop_p1.x = max(crop_p1.x+ rotated_offset, 0.0f);
        }

        cropimg = src(Rect(crop_p1, crop_p2));
    }

    mrz_rotatedRect.clear();
    dst = cropimg.clone();
}


void tdr_passport_post_process(Mat& src, int ret, OCR_RESULT& ocr_result, int isRefined, int sz1) {
    if(g_scan_none==1) return ;
    if (ret >= 0) {
        if (passport_ocr_result.result[0] == 'V' &&
              memcmp(passport_ocr_result.cntry1_result, "CHN", 3) == 0) {
              int rsp;
              if(isRefined==1)
                  rsp = get_visa_issue_info(src, ocr_result, true, sz1);
              else
                  rsp = get_visa_issue_info(src, ocr_result);
              setVisaCategory(ocr_result);
        }
        print_passport_result(ocr_result);
        cv::Mat faceimg;
        int count = getPassportFaceImage(src, faceimg);
        if (count == 0) {
            g_faceimg = faceimg.clone();
#ifdef TDR_TRACE
//if(isDebugShow()) imshow("face", g_faceimg);
#endif
        }
    }

    // 根据mrz区域取crop图像
    cropImage(src, g_cropimg, mrz_rotatedRect);
}

int tdr_passport_scan(Mat& src, OCR_RESULT& ocr_result) {
    double start = tdr_get_current_time();
    int sz1 = 0, sz2 = 0;
    int ret = passport_location(src, sz1, sz2);
    if (ret == 0) {
        ret = passportMatchOcr2(src, g_posRect, ocr_result);
        tdr_passport_post_process(src, ret, ocr_result, 0, 0);
    }
    double end = tdr_get_current_time();
    LOGD("tdr_passport_scan: %.3f ms", end - start);
    return ret;
}

int tdr_mixed_scan(Mat& src, OCR_RESULT& ocr_result) {
    double start = tdr_get_current_time();
    int sz1 = 0, sz2 = 0;
    int ret = passport_location(src, sz1, sz2);
    if (ret==-80)
    {
    }
    else if( ret< -20)
    {
    }
    else if (ret < 0) {
        ret = tdr_card_scan(src, ocr_result);
    } else if (ret == 0) {
        ret = passportMatchOcr2(src, g_posRect, ocr_result);
        tdr_passport_post_process(src, ret, ocr_result, 0, 0);
    } else {
        ret = -1;
    }
    double end = tdr_get_current_time();
    LOGD("tdr_mixed_scan: %.3f ms", end - start);
    return ret;
}

int tdr_mixed_scan_refine(Mat& src, OCR_RESULT& ocr_result) {

    double start = tdr_get_current_time();
    int sz1 = 0, sz2 = 0;
    cv::Mat line1_rgb, line2_rgb;
    int ret = passport_location3(src, sz1, sz2, line1_rgb, line2_rgb);
    if(ret==-80)
    {
    }
    else if( ret< -20)
    {
    }
    else if (ret < 0) {
      ret = tdr_card_scan(src, ocr_result);
    } else if (ret == 0) {
        ret=passportMatchOcr3(line1_rgb, line2_rgb, ocr_result, sz1, sz2);
        tdr_passport_post_process(src, ret, ocr_result, 1, sz1);
        //print_passport_result(ocr_result);
    } else {
        ret = -1;
    }
    double end = tdr_get_current_time();
    LOGD("tdr_mixed_scan_refine: %.3f ms", end - start);
    return ret;
}
